{{**
  * GPS Parse filters NMEA 0183 GPS data for specific sentence types (GGA, RMC,
  * etc.) and stores the sentences in the hub as null terminated strings.
  *
  * <p>
  *
  * This code is meant to work with any NMEA 0183 capable GPS, at any baud up to
  * 115200. The code is also clock speed independent: you can launch it at any
  * clock speed and the object calculates accordingly (but don't dynamically
  * change the clock speed after it is launched, and make sure that it's high
  * enough to handle the baud rate!)
  *
  * <p>
  *
  * This object works only with NMEA 0183 based devices. The driver waits until
  * it encounters a $ symbol (ASCII 36 decimal), then checks that the next two
  * characters are "G" and "P", and then matches a three character ID. This
  * three character ID is specified by the user so that specific sentences can
  * be allowed through the filter. If the sentence ID passes, then the entire
  * NMEA 0183 sentence (including the "$GP" and the ID) is recorded, followed by
  * a CR (ASCII 13 decimal) and a null terminator.
  * 
  * <p>
  * 
  * This code has been tested with the LS20031 GPS at 5Hz and 57600 baud.
  *
  * <p>
  *
  * Using this code:
  * 	- to adjust the buffer size, change NMEA_BUFFER_SIZE_CON 
  * 		(units are in NMEA strings, so each addition spot is 85 bytes)
  * 	- to adjust which strings are accepted:
  * 		I)  comment/uncomment/add the appropriate characters to NMEA_Watch
  * 		II) adjust MAX_NMEA_WATCH to match the number of strings to watch
  * 	- always include in your main program the following:
  * 		OBJ
  * 			GPS : "GPS_Parse.spin"
  * 		PUB Main
  * 			...
  * 			GPS.start(GPS_RX_PIN, GPS_BAUD)
  * 			...
  * 	- when you want to read the next GPS string, use
  * 			GPS.NMEA_str
  * 
  *
  * Example Useage:
  * 			debug.start(31, 30, 0, 115200)
  * 			GPS.start(6, 57600)
  * 			repeat
  * 				debug.str(GPS.NMEA_str) 'send all received strings to serial terminal	
  * 
  * 
  * @author  Cody Lewis <srlm@srlmproductions.com>
  * @version 1.0 Initial release
  * @since 8-25-2012
  *}}
CON
{{** adjust to change the number of strings in buffer. *}}
	NMEA_BUFFER_SIZE_CON = 16	
VAR
{{** cog flag/id *}}
	long  cog

{{** buffer head (write side) *}}
	long  rx_head

{{** buffer tail (read side) *}}
	long  rx_tail

{{**
  * buffer for nmea strings. Each string is 82 characters or less, guaranteed by
  * the NMEA 183 protocol.
  *}}
	byte NMEA_Buffer[NMEA_BUFFER_SIZE_CON * 85]


PUB start(rxpin, baudrate) : okay
{{**
  * Starts GPS serial driver in a new cog

  * @param rxpin    input receives signals from peripheral's TX pin
  * @param baudrate bits per second
  * @returns okay   returns false if no cog is available.
  *}}
	
	stop
	
	rx_head := 0
	rx_tail := 0

	NMEA_buffer_addr := @NMEA_Buffer
	NMEA_buffer_size := NMEA_BUFFER_SIZE_CON
	bitticks := clkfreq / baudrate
	rxmask := $1 << rxpin 'encode the RX pin
	
	okay := cog := cognew(@entry, @rx_head) + 1


PUB stop
{{**
  * Stops serial driver - frees a cog
  *}}
	if cog
		cogstop(cog~ - 1)
	longfill(@rx_head, 0, 2)

PUB NMEA_str : str_addr
{{**
  * This function will return the next received NMEA string. Note that it may
  * block while waiting for a string.
  *
  * @returns the address of a NMEA string (null terminated)
  *}}
	repeat while (str_addr := NMEA_str_check) < 0

PUB NMEA_str_flush 'WARNING: Untested (copied from FDS+)
{{**
  * Clear all the strings from the buffer without returning anything
  *}}
	repeat while NMEA_str_check => 0

PUB NMEA_str_check : str_addr
{{**
  * Check to see if a string is ready. Does not block.
  *
  * @returns -1 if no string received, string address if string is received
  *}}

	str_addr--
	if rx_tail <> rx_head
		str_addr := NMEA_buffer_addr + (rx_tail * 85)
		rx_tail := rx_tail + 1
		if rx_tail == NMEA_buffer_size
			rx_tail := 0

PUB NMEA_str_time(ms) : str_addr | t 'WARNING: Untested (copied from FDS+)
{{**
  * Wait for a specified number of milliseconds for a string to be received. If
  * no string is received during that time, the flag value of -1 is returned.
  * Otherwise, a string address is returned. Note that if there are strings 
  * already in the buffer when this function is called then it returns
  * immediately.
  *
  * @param   ms The number of milliseconds to wait.
  * @returns    -1 if no string received, address if string is received
  *}}

	t := cnt
	repeat until (str_addr := NMEA_str_check) => 0 or (cnt - t) / (clkfreq / 1000) > ms


DAT

'***********************************
'* Assembly language serial driver *
'***********************************

				org 0
entry	
'
' Receive
'
receive
				test    rxtxmode,#%001  wz    'wait for start bit on rx pin
				test    rxmask,ina      wc
	if_z_eq_c	jmp     #receive

				mov     rxbits,#9             'ready to receive byte
				mov     rxcnt,bitticks
				shr     rxcnt,#1
				add     rxcnt,cnt                          

:bit			add     rxcnt,bitticks        'ready next bit period

:wait
				mov     t1,rxcnt              'check if bit receive period done
				sub     t1,cnt
				cmps    t1,#0           wc
	if_nc		jmp     #:wait

				test    rxmask,ina      wc    'receive bit on rx pin
				rcr     rxdata,#1
				djnz    rxbits,#:bit

				shr     rxdata,#32-9          'justify and trim received byte
				and     rxdata,#$FF
				test    rxtxmode,#%001  wz    'if rx inverted, invert byte
	if_nz		xor     rxdata,#$FF



				
state_machine	
:switch			jmp		#:state_0			'Will be overwritten as appropriate

'-------------------------------------------	
:state_0 'Compare to $
				cmp		rxdata, #"$"	wz	'Compare to ASCII 's'
	if_z		movs	:switch, #:state_1	'transition to next state (1)
		
				'***************************
				'Do some housekeeping chores (here b/c of extra time between bytes)		
				'clear all the string counts
				movd	:state_0_loop, #NMEA_watch
				mov		t2, MAX_NMEA_WATCH			'Maximum number of strings headers

:state_0_loop	mov		0-0, #0						'Clear the match count
				add		:state_0_loop, movd_mask_4	'Increment destination to next match count variable
				djnz	t2, #:state_0_loop
		
				jmp		#receive
				
				
'-------------------------------------------
:state_1 'Compare to G
				cmp		rxdata, #"G"	wz	'Compare to ASCII
	if_z		movs	:switch, #:state_2	'transition to next state
	if_nz		movs	:switch, #:state_0	'character was not next, so back to state 0
	
				'***************************
				'Calculate the base hub address to store the next captured string at:
				'Note: this section included here b/c of
				'    a) there is spare time between bytes
				'    b) it can happen anywhere before state 5
				mov		string_tail, NMEA_buffer_addr
				
				rdlong	t2, par				    'get the head index in the hub
				add		t2, #1
				cmp		t2, NMEA_Buffer_size wc 'Is the head past the maximum?
	if_nc		mov		t2, #0					'if so, roll it over to zero
	
				cmp		t2, #0		wz
				
:state_1_inc
	if_nz		add		string_tail, #85
	if_nz		djnz	t2, #:state_1_inc
	
	
				jmp		#receive			

'-------------------------------------------
:state_2 'Compare to P
				cmp		rxdata, #"P"	wz	'Compare to ASCII
	if_z		movs	:switch, #:state_3	'transition to next state
	if_nz		movs	:switch, #:state_0	'character was not next, so back to state 0
				jmp		#receive			

'-------------------------------------------
:state_3 'Compare to first letter of string type
				movs	:switch, #:state_0			'presume that no character was found
				
				movs    :state_3_loop, #NMEA_watch	'the addr of the character of interest for that string
				add		:state_3_loop, #1			'get past the string counter, and to the FIRST digit
				movd	:state_3_store, #NMEA_watch
				mov		t3, MAX_NMEA_WATCH			't3 is the maximum different number of strings to have
				
:state_3_loop	cmp		rxdata, 0-0	wz				'Compare to ASCII letter
:state_3_store
	if_z		add		0-0, #1						'If the letter was found, increment string count
	if_z		movs	:switch, #:state_4			'transition to next state, if at least one is found
				add		:state_3_loop, #4
				add		:state_3_store, movd_mask_4
				djnz	t3, #:state_3_loop		
	
	
				jmp		#receive			

'-------------------------------------------
:state_4 'Compare to second letter of string type
				movs	:switch, #:state_0			'presume that no character was found
				
				movs    :state_4_loop, #NMEA_watch	'the character of interest for that string
				add		:state_4_loop, #2			'get past the string counter, and to the FIRST digit
				movd	:state_4_store, #NMEA_watch
				mov		t3, MAX_NMEA_WATCH			't3 is the maximum different number of strings to have
				
:state_4_loop	cmp		rxdata, 0-0	wz				'Compare to ASCII letter
:state_4_store
	if_z		add		0-0, #1						'If the letter was found, increment string count
	if_z		movs	:switch, #:state_5			'transition to next state, if at least one is found
				add		:state_4_loop, #4
				add		:state_4_store, movd_mask_4
				djnz	t3, #:state_4_loop		
	
	
				jmp		#receive			



'-------------------------------------------
:state_5 'Compare to third letter of string type
				movs	:switch, #:state_0			'presume that no character was found, REDUNDANT? IT MAY BE BELLOW TOO
				
				movs    :state_5_loop, #NMEA_watch	'the character of interest for that string
				add		:state_5_loop, #3			'get past the string counter, and to the FIRST digit
				movd	:state_5_store, #NMEA_watch
				mov		t3, MAX_NMEA_WATCH			't3 is the maximum different number of strings to have
				
:state_5_loop	cmp		rxdata, 0-0	wz				'Compare to ASCII letter
:state_5_store
	if_z		add		0-0, #1						'If the letter was found, increment string count
'	if_z		movs	:switch, #:state_6			'transition to next state, if at least one is found
				add		:state_5_loop, #4
				add		:state_5_store, movd_mask_4
				djnz	t3, #:state_5_loop		
	
	
	
				'***************************
				'Find the sentence type that matches (if any)
				mov		t1, #NMEA_watch
				movd	:state_5_cmp, t1
				mov		t2, MAX_NMEA_WATCH
				
:state_5_cmp	cmp		0-0, #3 wz				'Only passes if there are complete 3 letter matches
	if_nz		add		t1, #4					'+4 to get to next string
	if_nz		movd	:state_5_cmp, t1
	if_nz		djnz	t2, #:state_5_cmp
	
	if_nz		movs	:switch, #:state_0		'No matches found, return to start.

	if_z		movs	:switch, #:state_6		'Record the rest of the string if we have a match.
	if_z		add		t1, #1					'+1 to get to first character of string type
	if_z		call	#Write_header
				

				jmp		#receive		
		
				
'-------------------------------------------
:state_6 'Copy bytes to Hub until the end of the GPS string (indicated by CR (ASCII 13))
				cmp		rxdata, #10	wz		'Compare to ASCII LF
				
				'If it's going to be a case of writing this char,
				'	then check to make sure the string isn't too long.
				'	So we check *before* the cleanup code (in case it,
				'	fails, and we have to clean up).
'	if_nz		add		strlen, #1
'	if_nz		cmp		strlen, #85 wz
	
				'So, at this point if
				'	z = 0 then it's both not a CR, and it's not too long.
				'	z = 1 then it's either a CR, or it's too long.
				
				
				
				'***************************
				'Matches CR, so finished with string, clean up
'	if_z		mov		t1, #13             'Write a trailing CR to Hub NMEA string
'	if_z		wrbyte  t1, string_tail
'	if_z		add		string_tail, #1
	if_z		mov		t1, #0				'Null terminate the Hub string
	if_z		wrbyte  t1, string_tail
	
	if_z		movs	:switch, #:state_0	'All done with current NMEA string, time to wait for next string
	
	'increment head in Hub
	if_z		rdlong  t2,par                  'Read the current head index
	if_z		add     t2,#1                   'Increment the index
	
	if_z		cmp		t2, NMEA_Buffer_size wc 'Is the head past the maximum?
	if_z_and_nc	mov		t2, #0					'if so, roll it over to zero
	if_z		wrlong  t2,par                  'Save the index
	
'	if_z		mov		strlen, #0

	
				'***************************
				'Store char, wait for more
					
	if_nz		movs	:switch, #:state_6	'keep on recording characters
	if_nz		wrbyte  rxdata, string_tail
	if_nz		add		string_tail, #1

				jmp		#receive			
				
			
	
'-------------------------------------------
Write_header	'mov 	t2, #"$"
'				wrbyte  t2, string_tail
'				add		string_tail, #1

'				mov		t2, #"G"
'				wrbyte	t2, string_tail
'				add		string_tail, #1
'				
'				mov		t2, #"P"
'				wrbyte	t2, string_tail
'				add		string_tail, #1
'				
				movd	:a, t1
				add		t1, #1
:a				wrbyte	0-0, string_tail
				add		string_tail, #1

				movd	:b, t1
				add		t1, #1
:b				wrbyte	0-0, string_tail
				add		string_tail, #1

				movd	:c, t1
				nop
:c				wrbyte	0-0, string_tail
				add		string_tail, #1

Write_header_ret	ret

'-------------------------------------------
'strlen					long 	0

NMEA_Buffer_addr        long    0
NMEA_Buffer_size        long    0

rxmask                  long    0

rxtxmode                long    0
bitticks                long    0

MAX_NMEA_WATCH			long	5 'Make sure this matches the number of strings to watch!
NMEA_watch
s1						long	0,"G","S","V"
s4						long	0,"G","G","A"
s2						long	0,"R","M","C"
s3						long	0,"V","T","G"
s5						long	0,"G","L","L"
's6						long	0,"G","S","A"

movd_mask_4				long	%100_000000000 'adds 4_dec to the destination field


t1                      res     1
t2                      res     1
t3                      res     1


rxdata                  res     1
rxbits                  res     1
rxcnt                   res     1
rxcode                  res     1

string_tail             res     1

						fit 	496

{{
--------------------------------------------------------------------------------  
Copyright (c) 2012 Cody Lewis

Permission is hereby granted, free of charge, to any person obtaining a copy of

this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions: 

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software. 

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER

IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--------------------------------------------------------------------------------
}}

